# Copyright 2023-2024  Dom Sekotill <dom.sekotill@kodo.org.uk>

"""
Provides `Request`, a simple implementation of `konnect.curl.abc.RequestProtocol`
"""

from typing import IO

import pycurl


class Request:
	"""
	A simple implementation of `konnect.curl.abc.RequestProtocol`
	"""

	url: str
	destination: IO[bytes]

	def __init__(self, url: str, destination: IO[bytes]):
		self.url = url
		self.destination = destination

	def configure_handle(self, handle: pycurl.Curl, /) -> None:
		"""
		Set options on a `pycurl.Curl` instance to be used for this request
		"""
		handle.setopt(pycurl.URL, self.url)
		handle.setopt(pycurl.WRITEDATA, self.destination)
		handle.setopt(pycurl.CONNECTTIMEOUT_MS, 500)
		handle.setopt(pycurl.TIMEOUT_MS, 3000)

	def has_response(self) -> bool:
		"""
		Return whether calling `response()` will return a value or raise LookupError
		"""
		return False

	def response(self) -> None:
		"""
		Return a waiting response or raise LookupError if there is none

		See `has_response()` for checking for waiting responses.
		"""
		raise LookupError

	def completed(self) -> None:
		"""
		Indicate that Curl has completed processing the handle and return a final response
		"""



# Copyright 2023-2024  Dom Sekotill <dom.sekotill@kodo.org.uk>

from __future__ import annotations

from collections.abc import AsyncIterator
from collections.abc import Iterator
from contextlib import asynccontextmanager
from os import dup
from socket import SO_PROTOCOL
from socket import SO_TYPE
from socket import SOL_SOCKET
from socket import socket as Socket
from typing import Literal
from typing import Self
from typing import TypeAlias
from typing import TypeVar

import anyio
import pycurl
from anyio.abc import ObjectReceiveStream
from anyio.abc import ObjectSendStream

from ._enums import MILLISECONDS
from ._enums import SECONDS
from ._enums import SocketEvt
from ._enums import Time
from ._exceptions import CurlError
from .abc import RequestProtocol
from .scalars import Quantity

T = TypeVar("T")
Event: TypeAlias = tuple[Literal[SocketEvt.IN, SocketEvt.OUT], Socket]


class Multi:
	"""
	A wrapper around `pycurl.CurlMulti` to bind it with Anyio-supported event frameworks

	Users should treat this class as a black-box.  Only the `process()` method is available.
	There are no other methods or attributes they should attempt to call or modify as doing
	so will have undefined results.
	"""

	def __init__(self) -> None:
		self._handler = pycurl.CurlMulti()
		self._handler.setopt(pycurl.M_SOCKETFUNCTION, self._add_socket_evt)
		self._handler.setopt(pycurl.M_TIMERFUNCTION, self._add_timer_evt)
		self._io_events = dict[int, tuple[Socket, SocketEvt]]()
		self._deadline: Quantity[Time]|None = None
		self._perform_cond = anyio.Condition()
		self._governor_delegated = False
		self._completed = dict[pycurl.Curl, int]()
		self._requests = dict[RequestProtocol[object], pycurl.Curl]()

	def _add_socket_evt(self, what: int, socket: int, *_: object) -> None:
		# Callback registered with CURLMOPT_SOCKETFUNCTION, registers socket events the
		# transfer manager wants to be activated in response to.
		what: SocketEvt = SocketEvt(what)
		if what == SocketEvt.REMOVE:
			assert socket in self._io_events, f"file descriptor {socket} not in events"
			del self._io_events[socket]
		elif socket in self._io_events:
			socket_, __ = self._io_events[socket]
			self._io_events[socket] = socket_, what
		else:
			self._io_events[socket] = _ExternalSocket.from_fd(socket), what

	def _add_timer_evt(self, delay: int) -> None:
		# Callback registered with CURLMOPT_TIMERFUNCTION, registers when the transfer
		# manager next wants to be activated if no prior events occur, in milliseconds.
		self._deadline = \
			None if delay < 0 else \
			anyio.current_time() @ SECONDS + delay @ MILLISECONDS

	async def _single_event(self) -> int:
		# Await a single event and call pycurl.CurlMulti.socket_action to inform the handler
		# of it, then return the number of active transfers

		# Shortcut if no events are registered, or the only event is an immediate timeout
		if not self._io_events and self._deadline is None:
			_, running = self._handler.socket_action(pycurl.SOCKET_TIMEOUT, 0)
			return running

		# Start concurrent tasks awaiting each of the registered events, await a response
		# from the first task to wake and send one.
		async with anyio.create_task_group() as tasks, _make_evt_channel() as (sendchan, recvchan):
			for socket, evt in self._io_events.values():
				if SocketEvt.IN in evt:
					tasks.start_soon(_wait_readable, socket, sendchan)
				if SocketEvt.OUT in evt:
					tasks.start_soon(_wait_writable, socket, sendchan)
			if self._deadline is not None:
				tasks.start_soon(_wait_until, self._deadline, sendchan)
			resp = await recvchan.receive()
			tasks.cancel_scope.cancel()

		# Call pycurl.CurlMulti.socket_action() with details of the received event, and
		# return how many active handles remain.
		match resp:
			case None:
				_, running = self._handler.socket_action(pycurl.SOCKET_TIMEOUT, 0)
			case SocketEvt.IN, Socket() as socket:
				_, running = self._handler.socket_action(socket.fileno(), pycurl.CSELECT_IN)
			case SocketEvt.OUT, Socket() as socket:
				_, running = self._handler.socket_action(socket.fileno(), pycurl.CSELECT_OUT)  # type: ignore[unreachable]
		return running

	def _get_handle(self, request: RequestProtocol[object]) -> pycurl.Curl:
		try:
			return self._requests[request]
		except KeyError:
			handle = self._requests[request] = pycurl.Curl()
			request.configure_handle(handle)
			self._handler.add_handle(handle)
			return handle

	def _del_handle(self, request: RequestProtocol[object]) -> None:
		handle = self._requests.pop(request)
		self._handler.remove_handle(handle)

	def _yield_complete(self) -> Iterator[tuple[pycurl.Curl, int]]:
		# Convert pycurl.CurlMulti.info_read() output into tuples of (handle, code) and
		# iteratively yield them
		n_msgs = -1
		while n_msgs:
			# There is a bug in PyCurl affecting info_read; to work around it the
			# 'max_objects' argument must be AT LEAST the number of waiting messages, which
			# can be up to the total number of active handles. This makes the loop
			# superfluous but if the bug is fixed 'max_objects' could be replaced with
			# a static value to constrain memory usage and make it more predictable, and
			# remove the need to track the number of active handles.
			n_msgs, complete, failed = self._handler.info_read(len(self._requests))
			yield from ((handle, pycurl.E_OK) for handle in complete)
			yield from ((handle, res) for (handle, res, _) in failed)

	async def _govern_transfer(self, request: RequestProtocol[T], handle: pycurl.Curl) -> None:
		# Await _single_event() repeatedly until the wanted handle is completed.
		# Store all intermediate completed handles and notify interested tasks.
		remaining = -1
		while remaining:
			remaining = await self._single_event()
			self._completed.update(self._yield_complete())
			has_resp = request.has_response()
			if not has_resp and not self._completed:
				continue
			async with self._perform_cond:
				self._perform_cond.notify_all()
				# This check needs to be atomic with the notification
				# (more specifically, the _governor_delegated flag needs to be unset
				# atomically)
				if has_resp or handle in self._completed:
					self._governor_delegated = False
					return

		# SHOULD NOT fall off the end of the loop, but don't want it to run infinitely if
		# something goes wrong, so ensure it has a completion condition and raise
		# RuntimeError if it does complete
		raise RuntimeError("no response detected after all handles processed")

	async def process(self, request: RequestProtocol[T]) -> T:
		"""
		Perform a request as described by a Curl instance
		"""
		if request.has_response():
			return request.response()
		handle = self._get_handle(request)
		while handle not in self._completed:
			# If no task is governing the transfer manager, self-delegate the role to
			# ourselves and govern transfers until `handle` completes.
			if not self._governor_delegated:
				self._governor_delegated = True
				try:
					await self._govern_transfer(request, handle)
				finally:
					self._governor_delegated = False
			# Otherwise await a notification of completed handles
			else:
				async with self._perform_cond:
					await self._perform_cond.wait()
			if request.has_response():
				return request.response()
		match self._completed.pop(handle):  # noqa: R503
			case pycurl.E_OK:
				self._del_handle(request)
				return request.completed()
			case err:
				assert isinstance(err, int)  # nudge mypy
				self._del_handle(request)
				raise CurlError(err, handle.errstr())


class _ExternalSocket(Socket):
	"""
	A non-closing socket.socket subclass

	When instances are garbage-collected they do not close the underlying file descriptor.
	Used for sockets managed in a third-party library (libcurl in this case) which are
	passed as file descriptors.  The file descriptor MUST match, so duplication is not
	a possibility
	"""

	def __del__(self) -> None:
		return

	@classmethod
	def from_fd(cls, fd: int, /) -> Self:
		"""
		Create an instance from the file descriptor of an open socket
		"""
		# Create a temporary socket instance to get socket type and protocol (no low level
		# getsockopt()). FD must be duplicated as it will be closed when destroyed.
		socket = Socket(fileno=dup(fd))
		stype = socket.getsockopt(SOL_SOCKET, SO_TYPE)
		proto = socket.getsockopt(SOL_SOCKET, SO_PROTOCOL)
		return cls(socket.family, stype, proto, fd)


@asynccontextmanager
async def _make_evt_channel() -> AsyncIterator[tuple[ObjectSendStream[Event|None], ObjectReceiveStream[Event|None]]]:
	send, recv = anyio.create_memory_object_stream[Event|None](1)
	async with send, recv:
		yield send, recv


async def _wait_readable(socket: Socket, channel: ObjectSendStream[Event]) -> None:
	await anyio.wait_socket_readable(socket)
	await channel.send((SocketEvt.IN, socket))


async def _wait_writable(socket: Socket, channel: ObjectSendStream[Event]) -> None:
	await anyio.wait_socket_writable(socket)
	await channel.send((SocketEvt.OUT, socket))


async def _wait_until(time: Quantity[Time], channel: ObjectSendStream[None]) -> None:
	await anyio.sleep_until(time >> SECONDS)
	await channel.send(None)



# Copyright 2023-2024  Dom Sekotill <dom.sekotill@kodo.org.uk>

"""
Abstract protocols which may be implemented by users

`konnect.curl` provides simple implementations of these protocols which may be used directly
or subclassed by users.
"""

from typing import Protocol
from typing import TypeVar

from pycurl import Curl

T_co = TypeVar("T_co", covariant=True)


class RequestProtocol(Protocol[T_co]):
	"""
	Request classes that are passed to `Multi.process()` must implement this protocol
	"""

	def configure_handle(self, handle: Curl, /) -> None:
		"""
		Configure a Curl handle for the request by calling its `Curl.setopt()` method

		Users may wish to re-implement or wrap this method to override `Curl.setopt()`
		options.  The handle is guaranteed to be in a clean state with all options set to
		their defaults.
		"""
		...

	def has_response(self) -> bool:
		"""
		Return whether calling `response()` will return a value or raise `LookupError`
		"""
		...

	def response(self) -> T_co:
		"""
		Return a waiting response or raise `LookupError` if there is none

		See `has_response()` for checking for waiting responses.
		"""
		...

	def completed(self) -> T_co:
		"""
		Indicate that Curl has completed processing the handle and return a final response
		"""
		...



# Copyright 2023  Dom Sekotill <dom.sekotill@kodo.org.uk>

from ._enums import MILLISECONDS as MILLISECONDS
from ._enums import SECONDS as SECONDS
from ._enums import Time as Time
from ._exceptions import CurlError as CurlError
from ._multi import Multi as Multi
from .requests import Request as Request



# Copyright 2023-2024  Dom Sekotill <dom.sekotill@kodo.org.uk>

"""
Scalar quantities, which have a magnitude and a unit

Defining Units
--------------

Unit types for `Quantity` are created by subclassing the `QuantityUnit` `enum` base:

>>> class Time(QuantityUnit):
...     MILLISECONDS = 1
...     SECONDS = 1000
...     MINUTES = 60 * SECONDS

>>> class Distance(QuantityUnit):
...     MILLIMETERS = 1
...     CENTIMETERS = 10 * MILLIMETERS
...     METERS = 100 * CENTIMETERS
...     KILOMETERS = 1000 * METERS
...     INCH = 24 * MILLIMETERS
...     HALF_INCH = INCH // 2  # Hey! Hands off
...     QUARTER_INCH = INCH // 4

The enum members form units of relative size to each other. They MUST be integers so
typically the smallest (highest precision) unit is `1` and the others are some multiple of
it. Note that the designated 'unit' can be changed without breaking dependant code (as long
as the code is using the quantities right).  In this case the unitary value is 0.5mm:

>>> class Distance(QuantityUnit):
...     MILLIMETERS = 2  # Scaled to allow SIXTEENTH_INCH to be an integer
...     CENTIMETERS = 10 * MILLIMETERS
...     METERS = 100 * CENTIMETERS
...     KILOMETERS = 1000 * METERS
...     INCH = 24 * MILLIMETERS
...     # ...
...     SIXTEENTH_INCH = INCH // 16  # 1/16″ is 1.5mm


Creating Scalar Quantities
--------------------------

A scalar quantity can be created using the matrix multiplication operator "@" with a
scalar unit, e.g. 2 seconds:

>>> quantity: Quantity[Time] = 2 @ Time.SECONDS

Units of the same type relate to one another as you would expect (parentheses for clarity):

>>> assert (2 @ Time.SECONDS) == (2000 @ Time.MILLISECONDS)

Quantities may also be multiplied (and divided) by unitless values:

>>> ONE_SECOND = 1 @ Time.SECONDS
>>> assert (ONE_SECOND / 2) == (500 @ Time.MILLISECONDS)

Note that quantities are really just integers which, at runtime, have no additional
information attached to them. This means that Python will happily accept any `Quantity`
wherever it would accept an integer; however static type checkers such as MyPy will complain
about it, which is good as it is almost certainly a mistake to attempt to, for example, sum
time and distance quantities, or sum a quantity with an arbitrary value:

>>> meaningless_value = (2 @ Time.SECONDS) + (10 @ Distance.MILLIMETERS)

>>> # Depending on the declaration of Distance, 100 here could be 100mm or 100/16″,
... # or anything else...
... unreliable_value = (2 @ Distance.METERS) + 100

Multiplying quantities with other quantities, even of the same type, would produce
a different unit, which is not supported.  (However, it is not inconceivable that it could
be supported in the future.) The following will also fail static type checks:

>>> area = (2 @ Distance.METERS) * (2 @ Distance.METERS)  # 4.0m²
>>> speed = (100 @ Distance.METERS) / (1 @ Time.SECOND)  # 100m/s


Using Scalar Quantities
-----------------------

At some point quantities will need to be passed through an interface of some sort where the
unit information will be lost.  Such interfaces will define a single unit they accept; for
instance `time.sleep()` requires an argument in seconds.  Upon reaching such an interface,
quantities can be stripped of their scalar types and converted to the required unit with the
right bit-shift operator ">>" or integer division operator "//":

>>> delay = 2 @ Time.MINUTES
>>> ...
>>> time.sleep(delay >> Time.SECONDS)

With the ">>" operator the type of the resulting value is always a `float` and
will only be precise up to the highest precision unit for a defined `QuantityUnit` type (the
unit with a magnitude of `1`, which need not be explicitly defined).

With the "//" operator the resulting type will be `int`, with whatever loss of precision
that implies.


Choice of Operators
-------------------

The operators for constructing ("@") and deconstructing (">>") quantities may seem
a bit odd, given that what they actually do is multiply and divide the values.
They were chosen to be visually distinct from other multiplication and division operations
on quantities and scalar units.

The matrix multiplication operator therefore replaces the scalar multiplication operator,
while the shift operator, which looks arrow-like, is used to convert to the indicated
unit.

>>> # As in ...
... # quantity  (converted to)  units
... time              >>        Time.SECONDS
"""

from __future__ import annotations

import enum
from typing import TYPE_CHECKING
from typing import Generic
from typing import Self
from typing import TypeVar
from typing import overload

U = TypeVar("U", bound="QuantityUnit")


class Quantity(int, Generic[U]):
	"""
	A scalar quantity of a given type 'U'
	"""

	if TYPE_CHECKING:

		@overload
		def __add__(self, other: Quantity[U]) -> Quantity[U]: ...

		@overload
		def __add__(self, other: int) -> int: ...

		def __add__(self, other: int) -> int:
			"""
			Adding two quantities creates a new quantity
			"""
			...

		@overload
		def __sub__(self, other: Quantity[U]) -> Quantity[U]: ...

		@overload
		def __sub__(self, other: int) -> int: ...

		def __sub__(self, other: int) -> int:
			"""
			Subtracting a quantity from another creates a new quantity
			"""
			...

		def __mul__(self, other: int|float) -> Quantity[U]:
			"""
			Quantities can be multiplied by unitless values to produce a new quantity
			"""
			...

		def __truediv__(self, other: int|float) -> Quantity[U]:
			"""
			Quantities can be divided by unitless values to produce a new quantity
			"""
			...


class QuantityUnit(enum.Enum):
	"""
	Enum base class for units
	"""

	if TYPE_CHECKING:
		@property
		def value(self) -> int: ...  # noqa: D102

	def __rmatmul__(self, scalar: float|int) -> Quantity[Self]:
		return Quantity(self.value * scalar)

	def __rrshift__(self, quantity: Quantity[Self]) -> float:
		return quantity / self.value

	def __rfloordiv__(self, quantity: Quantity[Self]) -> int:
		return quantity // self.value



# Copyright 2023  Dom Sekotill <dom.sekotill@kodo.org.uk>


class CurlError(Exception):
	"""
	An error exception for Curl transfers

	The value of 'code' is a Curl `CURLE_*` value, for instance `CURLE_COULDNT_CONNECT`.
	The value of 'msg' is an error message returned by `curl_easy_strerror` with 'code'
	passed as an argument.
	"""

	args: tuple[int, str]

	def __init__(self, code: int, msg: str):
		super().__init__(code, msg)

	def __str__(self) -> str:
		return f"Curl error: ({self.args[0]}) {self.args[1]}"

	@property
	def code(self) -> int:
		return self.args[0]

	@property
	def msg(self) -> str:
		return self.args[1]



# Copyright 2023  Dom Sekotill <dom.sekotill@kodo.org.uk>

"""
Enum classes for other modules
"""

import enum
from typing import Final

import pycurl

from .scalars import QuantityUnit


class SocketEvt(enum.Flag):
	"""
	Provides the `CURL_POLL_*` values from libcurl as a Python enum

	Member names have the 'CURL_POLL_' portion of the name removed.
	"""

	IN = pycurl.POLL_IN
	OUT = pycurl.POLL_OUT
	INOUT = pycurl.POLL_INOUT
	REMOVE = pycurl.POLL_REMOVE


class Time(QuantityUnit):
	"""
	Units for time intervals
	"""

	# Base times on milliseconds as thats the greatest precision used by curl.
	# This can be transparently changed at any time to a greater precision base.

	MILLISECONDS = 1
	SECONDS = 1000


MILLISECONDS: Final = Time.MILLISECONDS
SECONDS: Final = Time.SECONDS
